"""AAS Standard parser Package."""

import importlib.metadata
from datetime import UTC, datetime

from aas_standard_parser import (
    aas_parser,
    aid_parser,
    aimc_parser,
    collection_helpers,
    descriptor_json_helper,
    reference_helpers,
    submodel_json_helper,
    submodel_parser,
    utils,
)
from aas_standard_parser.aid_parser import AIDParser
from aas_standard_parser.version_check import check_for_update

__copyright__ = f"Copyright (C) {datetime.now(tz=UTC).year} Fluid 4.0. All rights reserved."
__author__ = "Daniel Klein, Celina Adelhardt, Tom Gneuß"

try:
    __license__ = "MIT"
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0-dev"

__project__ = "aas-standard-parser"
__package__ = "aas-standard-parser"


check_for_update()

__all__ = [
    "AIDParser",
    "aas_parser",
    "aid_parser",
    "aimc_parser",
    "collection_helpers",
    "descriptor_json_helper",
    "reference_helpers",
    "submodel_json_helper",
    "submodel_parser",
    "utils",
]



"""Helper functions for working with ModelReference objects in the AAS standard parser."""

from basyx.aas.model import ModelReference


def construct_id_short_path_from_reference(reference: ModelReference) -> str:
    """Constructs an idShort path from the given ModelReference.

    :param reference: The ModelReference to construct the idShort path from.
    :return: The constructed idShort path as a string.
    """
    id_short_path: str = ""

    # start from the second Key and omit the Identifiable at the beginning of the list
    for key in reference.key[1:]:
        id_short_path += key.value + "."

    # get rid of the trailing dot
    return id_short_path[:-1]


def get_values_from_keys(reference: ModelReference) -> list[str]:
    """Returns the values from all keys in reference as list.

    :param reference: reference to extract values from
    :return: list of values from all keys in the reference
    """
    return [key.value for key in reference.key]


def get_value_from_key_at_index(reference: ModelReference, index: int) -> str:
    """Returns the value from the key at the given index in reference.

    :param reference: reference to extract value from
    :param index: index of the key to get the value from
    :return: value from the key at the given index
    """
    return reference.key[index].value



"""Module for parsing AAS."""

import logging

from basyx.aas import model

logger = logging.getLogger(__name__)


def get_submodel_ids(shell: model.AssetAdministrationShell) -> list[str]:
    """Get all IDs from the submodels referenced in the given AAS.

    :param shell: The Asset Administration Shell to extract submodel IDs from.
    :return: A list of submodel IDs referenced in the AAS.
    """
    submodel_ids = []
    for submodel in shell.submodel:
        if len(submodel.key) < 1 or submodel.key[0].type != model.KeyTypes.SUBMODEL:
            logger.warning(f"Submodel reference {submodel} does not start with SUBMODEL key type.")
            continue

        submodel_ids.append(submodel.key[0].value)

    return submodel_ids



import json
from dataclasses import field

import basyx.aas.adapter.json
from basyx.aas import model


class ReferenceProperties:
    """Class representing properties of a reference in the mapping configuration."""

    reference: model.ExternalReference = field(metadata={"description": "Reference to the property in the submodel."})
    submodel_id: str = field(metadata={"description": "Identifier of the submodel used by the reference."})
    property_name: str = field(metadata={"description": "Name of the mapped property."})
    parent_path: list[str] = field(metadata={"description": "List of idShorts representing the parent path of the reference."})


class SourceSinkRelation:
    """Class representing a source-sink relation in the mapping configuration."""

    source_properties: ReferenceProperties = field(metadata={"description": "Properties of the source reference."})
    sink_properties: ReferenceProperties = field(metadata={"description": "Properties of the sink reference."})

    def source_reference_as_dict(self) -> dict:
        """Convert the source reference to a dictionary.

        :return: The source reference as a dictionary.
        """
        dict_string = json.dumps(self.source_properties.reference, cls=basyx.aas.adapter.json.AASToJsonEncoder)
        dict_string = dict_string.replace("GlobalReference", "Submodel").replace("FragmentReference", "SubmodelElementCollection")
        return json.loads(dict_string)

    def sink_reference_as_dict(self) -> dict:
        """Convert the sink reference to a dictionary.

        :return: The sink reference as a dictionary.
        """
        return json.loads(json.dumps(self.sink_properties.reference, cls=basyx.aas.adapter.json.AASToJsonEncoder))


class MappingConfiguration:
    """Class representing a mapping configuration."""

    interface_reference: model.ReferenceElement = field(metadata={"description": "Reference to the interface in the AID submodel."})
    aid_submodel_id: str = field(metadata={"description": "Identifier of the AID submodel used by the interface reference."})
    source_sink_relations: list[SourceSinkRelation] = field(metadata={"description": "List of source-sink relations in the mapping configuration."})


class MappingConfigurations:
    """Class representing mapping configurations from AIMC submodel."""

    configurations: list[MappingConfiguration] = field(metadata={"description": "List of mapping configurations."})
    aid_submodel_ids: list[str] = field(metadata={"description": "List of AID submodel IDs used in the mapping configurations."})



"""Module for parsing submodels."""

import logging

from basyx.aas import model

logger = logging.getLogger(__name__)


def get_submodel_element_by_id_short_path(submodel: model.Submodel, id_short_path: str) -> model.SubmodelElement:
    """Retrieve a specific submodel element from the submodel at a specific idShort path.

    :param submodel: The submodel to search within.
    :param id_short_path: IdShort path to the submodel element (dot-separated), e.g., "Element1.Element2[0].Element3".
    :return: The found submodel element or None if not found.
    """
    # Split the path by '.' and traverse the structure
    parts = id_short_path.split(".")
    current_elements = submodel.submodel_element
    part_index = 0
    for part in parts:
        part_index += 1
        # Handle indexed access like "Element[0]" for SubmodelElementLists
        if "[" in part and "]" in part:
            # Split SubmodelElementList name and index
            base, idx = part[:-1].split("[")
            idx = int(idx)
            # Find the SubmodelElementList in the current elements
            submodel_element = next((el for el in current_elements if el.id_short == base), None)

            if not submodel_element or not (isinstance(submodel_element, (model.SubmodelElementList, model.SubmodelElementCollection))):
                logger.debug(f"Submodel element '{base}' not found or is not a collection/list in current {current_elements}.")
                return None

            # Check if index is within range
            if idx >= len(submodel_element.value):
                logger.debug(f"Index '{idx}' out of range for element '{base}' with length {len(submodel_element.value)}.")
                return None

            # get the element by its index from SubmodelElementList
            submodel_element = submodel_element.value[idx]

        else:
            # Find the SubmodelElement in the current SubmodelElementCollection
            submodel_element = next((el for el in current_elements if el.id_short == part), None)

        if not submodel_element:
            logger.debug(f"Submodel element '{part}' not found in current {current_elements}.")
            return None

        # If we've reached the last part, return the found element
        if part_index == len(parts):
            return submodel_element

        # If the found element is a collection or list, continue traversing
        if isinstance(submodel_element, (model.SubmodelElementCollection, model.SubmodelElementList)):
            current_elements = submodel_element.value
        else:
            return submodel_element

    return submodel_element


def get_semantic_id_value(submodel: model.Submodel, index: int = 0) -> str | None:
    """Get the semantic ID from a submodel.

    :param submodel: The submodel to extract the semantic ID from.
    :param index: The index of the semantic ID key to retrieve.
    :return: The semantic ID string if found, otherwise None.
    """
    if submodel.semantic_id is None or index >= len(submodel.semantic_id.key):
        logger.warning(f"No semantic ID found for submodel {submodel.id_short}")
        return None

    return submodel.semantic_id.key[index].value



"""Module for parsing descriptor JSON data."""

import logging

from aas_standard_parser.classes.descriptor_json_helper_classes import DescriptorData, EndPointHrefData
from aas_standard_parser.utils import decode_base_64, encode_base_64

logger = logging.getLogger(__name__)


def parse_descriptor(descriptor_data: dict) -> DescriptorData | None:
    """Parse the descriptor data into a DescriptorData object.

    :param descriptor_data: The descriptor data containing endpoints.
    :return: A DescriptorData object containing parsed descriptor information.
    """
    identifier_encoded = descriptor_data.get("id", "")

    if not identifier_encoded:
        logger.warning(f"Descriptor data missing 'id' field: {descriptor_data}")
        return None

    descriptor = DescriptorData(identifier_encoded)
    descriptor.endpoints = descriptor_data.get("endpoints", [])
    descriptor.description = descriptor_data.get("description", {})
    descriptor.display_name = descriptor_data.get("displayName", {})
    descriptor.assetKind = descriptor_data.get("assetKind", "")
    descriptor.id_short = descriptor_data.get("idShort", "")
    descriptor.semantic_id = descriptor_data.get("semanticId", {})
    descriptor.supplementary_semantic_ids = descriptor_data.get("supplementarySemanticIds", [])
    descriptor.identifier = encode_base_64(identifier_encoded)

    return descriptor


def get_endpoints(descriptor_data: dict) -> list[dict]:
    """Get all endpoints from a descriptor.

    :param descriptor_data: The descriptor data containing endpoints.
    :return: A list of endpoint dictionaries extracted from the descriptor.
    """
    return descriptor_data.get("endpoints", [])


def get_endpoint_by_index(descriptor_data: dict, endpoint_index: int = 0) -> dict | None:
    """Get the endpoint from a descriptor's endpoints by index.

    :param descriptor_data: The descriptor data containing endpoints.
    :param endpoint_index: The index of the endpoint to extract.
    :return: The endpoint dictionary if found, otherwise None.
    """
    endpoints = get_endpoints(descriptor_data)

    if not endpoints or len(endpoints) == 0:
        logger.warning(f"No endpoints found in descriptor {descriptor_data}")
        return None

    if endpoint_index >= len(endpoints):
        logger.warning(f"Endpoint index {endpoint_index} out of range for descriptor {descriptor_data}")
        return None

    return endpoints[endpoint_index]


def get_endpoint_protocol_information_by_index(descriptor_data: dict, endpoint_index: int = 0) -> dict | None:
    """Get the protocol information from a descriptor's endpoints by index.

    :param descriptor_data: The descriptor data containing endpoints.
    :param endpoint_index: The index of the endpoint to extract the protocol information from.
    :return: The protocol information dictionary if found, otherwise None.
    """
    endpoint = get_endpoint_by_index(descriptor_data, endpoint_index)

    if endpoint is None:
        logger.warning(f"Endpoint at index {endpoint_index} not found in descriptor {descriptor_data}")
        return None

    return endpoint.get("protocolInformation", {})


def get_endpoint_href_by_index(descriptor_data: dict, endpoint_index: int = 0) -> str | None:
    """Get the href from a descriptor's endpoints.

    :param descriptor_data: The descriptor data containing endpoints.
    :param endpoint_index: The index of the endpoint to extract the href from.
    :return: The href string if found, otherwise None.
    """
    protocol_info = get_endpoint_protocol_information_by_index(descriptor_data, endpoint_index)

    if not protocol_info:
        logger.warning(f"No protocol information found for endpoint at index {endpoint_index} in descriptor {descriptor_data}")
        return None

    return protocol_info.get("href", "")


def get_endpoint_hrefs(descriptor_data: dict) -> list[str]:
    """Get all hrefs from a descriptor's endpoints.

    :param descriptor_data: The descriptor data containing endpoints.
    :return: A list of href strings extracted from the endpoints.
    """
    endpoints = get_endpoints(descriptor_data)

    return [endpoint.get("protocolInformation", {}).get("href", "") for endpoint in endpoints]


def parse_endpoint_href(href: str) -> EndPointHrefData | None:
    """Parse the endpoint href into its components.

    :param href: The href string to parse.
    :return: An EndPointHrefData object containing parsed components.
    """
    if "shells/" not in href and "submodels/" not in href:
        logger.warning(f"Invalid href format: {href}")
        return None

    href_data = EndPointHrefData(href)

    split_str = ""
    if "/shells/" in href:
        split_str = "/shells/"
        href_data.tag = split_str.replace("/", "")
    elif "/submodels/" in href:
        split_str = "/submodels/"
        href_data.tag = split_str.replace("/", "")

    base_url: str = href.split(split_str, maxsplit=1)[0]
    identifier: str = href.split(split_str)[1]

    href_data.base_url = base_url
    href_data.identifier = identifier

    try:
        href_data.identifier_decoded = decode_base_64(identifier)
    except Exception as e:
        logger.error(f"Failed to decode identifier '{identifier}' from href '{href}': {e}")

    return href_data



import pytest
from basyx.aas import model

from aas_standard_parser.classes.aimc_parser_classes import MappingConfiguration, SourceSinkRelation, ReferenceProperties
import aas_standard_parser.aimc_parser as aimc_parser
from aas_standard_parser.utils import create_submodel_from_file


@pytest.fixture(scope="module")
def aimc_submodel() -> model.Property:
    # create a Submodel
    return create_submodel_from_file("tests/test_data/aimc_submodel.json")


def test_001_get_mapping_configuration_root_element(aimc_submodel: model.submodel):
    root_element = aimc_parser.get_mapping_configuration_root_element(aimc_submodel)

    assert root_element is not None
    assert root_element.id_short == "MappingConfigurations"


def test_002_get_mapping_configuration_elements(aimc_submodel: model.submodel):
    configuration_elements = aimc_parser.get_mapping_configuration_elements(
        aimc_submodel
    )

    assert configuration_elements is not None
    assert len(configuration_elements) == 1

    configuration_element = configuration_elements[0]
    assert isinstance(configuration_element, model.SubmodelElementCollection)

def test_003_parse_mapping_configuration_element(aimc_submodel: model.submodel):
    configuration_elements = aimc_parser.get_mapping_configuration_elements(
        aimc_submodel
    )

    configuration_element = configuration_elements[0]
    configuration = aimc_parser.parse_mapping_configuration_element(configuration_element)

    assert configuration is not None
    _check_interface_ref(configuration)
    _check_relations(configuration)

def test_004_parse_mapping_configurations(aimc_submodel: model.submodel):
    mapping_configurations = aimc_parser.parse_mapping_configurations(aimc_submodel)

    assert mapping_configurations is not None
    assert len(mapping_configurations.configurations) == 1
    assert len(mapping_configurations.aid_submodel_ids) == 1

    configuration = mapping_configurations.configurations[0]
    assert configuration.aid_submodel_id is not None
    assert configuration.aid_submodel_id in mapping_configurations.aid_submodel_ids
    assert configuration.aid_submodel_id == "https://fluid40.de/ids/sm/4757_4856_8464_1441"

    _check_interface_ref(configuration)
    _check_relations(configuration)

def _check_interface_ref(configuration: MappingConfiguration):
    assert configuration.interface_reference is not None

    interface_reference = configuration.interface_reference
    assert isinstance(interface_reference, model.ReferenceElement)
    assert interface_reference.id_short == "InterfaceReference"

    value = interface_reference.value
    assert isinstance(value, model.Reference)
    assert len(value.key) == 1

    key = value.key[0]
    assert key is not None
    assert key.value == "https://fluid40.de/ids/sm/4757_4856_8464_1442/Interface_MQTT"


def _check_relations(configuration: MappingConfiguration):
    assert len(configuration.source_sink_relations) > 0

    relation = configuration.source_sink_relations[0]
    assert relation is not None

    assert relation.source_properties is not None
    assert relation.source_properties.submodel_id == configuration.aid_submodel_id

    assert relation.sink_properties is not None

    _check_relations_source(relation.source_properties)
    _check_relations_sink(relation.sink_properties)
    _check_relation_methods(relation)

def _check_relations_source(reference_properties: ReferenceProperties):
    assert reference_properties is not None

    assert isinstance(reference_properties.reference, model.ExternalReference)

    assert reference_properties.property_name is not None
    assert reference_properties.property_name == "HandlingC"

    assert reference_properties.parent_path is not None
    assert len(reference_properties.parent_path) == 5
    assert reference_properties.parent_path[3] == "axes_position"

    assert reference_properties.submodel_id is not None
    assert reference_properties.submodel_id == reference_properties.reference.key[0].value

def _check_relations_sink(reference_properties: ReferenceProperties):
    assert reference_properties is not None

    assert isinstance(reference_properties.reference, model.ExternalReference)

    assert reference_properties.property_name is not None
    assert reference_properties.property_name == "HandlingC"

    assert reference_properties.parent_path is not None
    assert len(reference_properties.parent_path) == 2
    assert reference_properties.parent_path[1] == "AxesPosition"

    assert reference_properties.submodel_id is not None
    assert reference_properties.submodel_id == reference_properties.reference.key[0].value

def _check_relation_methods(relation: SourceSinkRelation):
    # test to_json methods
    source_json = relation.source_reference_as_dict()
    assert isinstance(source_json, dict)
    assert source_json is not None
    assert "type" in source_json

    sink_json = relation.sink_reference_as_dict()
    assert isinstance(sink_json, dict)
    assert sink_json is not None
    assert "type" in sink_json



"""Helper classes for descriptor JSON parsing."""


class EndPointHrefData:
    """Class to represent an endpoint href data structure."""

    def __init__(self, href: str):
        """Initialize the EndPointHrefData with the given href."""
        self.href: str = href
        self.base_url: str = ""
        self.identifier: str = ""
        self.identifier_decoded: str = ""
        self.tag: str = ""


class DescriptorData:
    """Class to represent the descriptor data structure."""

    def __init__(self, identifier_decoded: str):
        """Initialize the DescriptorData with the given descriptor data."""
        self.identifier_decoded: str = identifier_decoded
        self.identifier: str = ""
        self.endpoints: list[dict] = []
        self.description: dict = {}
        self.display_name: dict = {}
        self.assetKind = ""
        self.id_short = ""
        self.semantic_id = ""
        self.supplementary_semantic_ids: list[str] = []



"""Parser for Mapping Configurations in AIMC Submodel."""

import logging

from basyx.aas import model

import aas_standard_parser.collection_helpers as ch
from aas_standard_parser.classes.aimc_parser_classes import MappingConfiguration, MappingConfigurations, ReferenceProperties, SourceSinkRelation

logger = logging.getLogger(__name__)


def get_mapping_configuration_root_element(aimc_submodel: model.Submodel) -> model.SubmodelElementCollection | None:
    """Get the mapping configuration root submodel element collection from the AIMC submodel.

    :param aimc_submodel: The AIMC submodel to extract the mapping configuration root from.
    :return: The mapping configuration root submodel element collection or None if not found.
    """
    # check if AIMC submodel is None
    if aimc_submodel is None:
        logger.error("AIMC submodel is None.")
        return None

    # get 'MappingConfigurations' element list by its semantic ID
    mc_element = ch.find_by_in_semantic_id(aimc_submodel.submodel_element, "idta/AssetInterfacesMappingConfiguration/1/0/MappingConfigurations")
    if mc_element is None:
        logger.error("'MappingConfigurations' element list not found in AIMC submodel.")

    return mc_element


def get_mapping_configuration_elements(aimc_submodel: model.Submodel) -> list[model.SubmodelElementCollection] | None:
    """Get all mapping configurations from the AIMC submodel.

    :param aimc_submodel: The AIMC submodel to extract mapping configurations from.
    :return: A dictionary containing all mapping configurations.
    """
    # check if AIMC submodel is None
    if aimc_submodel is None:
        logger.error("AIMC submodel is None.")
        return None

    # get mapping configuration root element
    root_element = get_mapping_configuration_root_element(aimc_submodel)
    if root_element is None:
        return None

    # find all 'MappingConfiguration' elements by their semantic ID
    mapping_configurations = ch.find_all_by_in_semantic_id(root_element.value, "idta/AssetInterfacesMappingConfiguration/1/0/MappingConfiguration")

    logger.debug(f"Found {len(mapping_configurations)} mapping configuration elements in AIMC submodel.")

    return mapping_configurations


def parse_mapping_configurations(aimc_submodel: model.Submodel) -> MappingConfigurations | None:
    """Parse all mapping configurations in the AIMC submodel.

    :param aimc_submodel: The AIMC submodel to parse mapping configurations from.
    :return: A list of parsed mapping configurations or None if parsing failed.
    """
    logger.info("Parse mapping configurations from AIMC submodel.")

    mapping_configurations: list[MappingConfiguration] = []

    # get all mapping configuration elements
    mapping_configurations_elements = get_mapping_configuration_elements(aimc_submodel)
    if mapping_configurations_elements is None:
        logger.error("No mapping configuration elements found in AIMC submodel.")
        return None

    # parse each mapping configuration element
    for mc_element in mapping_configurations_elements:
        mc = parse_mapping_configuration_element(mc_element)
        if mc is not None:
            mapping_configurations.append(mc)

    logger.debug(f"Parsed {len(mapping_configurations)} mapping configurations.")

    mcs = MappingConfigurations()
    mcs.configurations = mapping_configurations
    # add all unique AID submodel IDs from all mapping configurations
    mcs.aid_submodel_ids = list({mc.aid_submodel_id for mc in mapping_configurations})

    logger.debug(f"Found {len(mcs.aid_submodel_ids)} unique AID submodel IDs in mapping configurations.")
    logger.debug(f"Found {len(mcs.configurations)} mapping configurations in AIMC submodel.")

    return mcs


def parse_mapping_configuration_element(
    mapping_configuration_element: model.SubmodelElementCollection,
) -> MappingConfiguration | None:
    """Parse a mapping configuration element.

    :param mapping_configuration_element: The mapping configuration element to parse.
    :return: The parsed mapping configuration or None if parsing failed.
    """
    if mapping_configuration_element is None:
        logger.error("Mapping configuration element is None.")
        return None

    logger.debug(f"Parse mapping configuration '{mapping_configuration_element}'")

    # get interface reference element
    interface_reference = _get_interface_reference_element(mapping_configuration_element)
    if interface_reference is None:
        return None

    source_sink_relations = _generate_source_sink_relations(mapping_configuration_element)

    if len(source_sink_relations) == 0:
        logger.error(f"No source-sink relations found in mapping configuration '{mapping_configuration_element.id_short}'.")
        return None

    # check if all relations have the same AID submodel
    aid_submodel_ids = list({source_sink_relation.source_properties.submodel_id for source_sink_relation in source_sink_relations})

    if len(aid_submodel_ids) != 1:
        logger.error(
            f"Multiple AID submodel IDs found in mapping configuration '{mapping_configuration_element.id_short}': {aid_submodel_ids}. Expected exactly one AID submodel ID."
        )
        return None

    mc = MappingConfiguration()
    mc.interface_reference = interface_reference
    mc.source_sink_relations = source_sink_relations
    # add all unique AID submodel IDs from source-sink relations
    mc.aid_submodel_id = aid_submodel_ids[0]
    return mc


def _get_interface_reference_element(
    mapping_configuration_element: model.SubmodelElementCollection,
) -> model.ReferenceElement | None:
    """Get the interface reference ID from the mapping configuration element.

    :param mapping_configuration_element: The mapping configuration element to extract the interface reference ID from.
    :return: The interface reference ID or None if not found.
    """
    logger.debug(f"Get 'InterfaceReference' from mapping configuration '{mapping_configuration_element}'.")

    interface_ref: model.ReferenceElement = ch.find_by_in_semantic_id(
        mapping_configuration_element, "idta/AssetInterfacesMappingConfiguration/1/0/InterfaceReference"
    )

    if interface_ref is None or not isinstance(interface_ref, model.ReferenceElement):
        logger.error(f"'InterfaceReference' not found in mapping configuration '{mapping_configuration_element.id_short}'.")
        return None

    if interface_ref.value is None or len(interface_ref.value.key) == 0:
        logger.error(f"'InterfaceReference' has no value in mapping configuration '{mapping_configuration_element.id_short}'.")
        return None

    return interface_ref


def _generate_source_sink_relations(mapping_configuration_element: model.SubmodelElementCollection) -> list[SourceSinkRelation]:
    source_sink_relations: list[SourceSinkRelation] = []

    logger.debug(f"Get 'MappingSourceSinkRelations' from mapping configuration '{mapping_configuration_element}'.")

    relations_list: model.SubmodelElementList = ch.find_by_in_semantic_id(
        mapping_configuration_element, "/idta/AssetInterfacesMappingConfiguration/1/0/MappingSourceSinkRelations"
    )

    if relations_list is None or not isinstance(relations_list, model.SubmodelElementList):
        logger.error(f"'MappingSourceSinkRelations' not found in mapping configuration '{mapping_configuration_element.id_short}'.")
        return source_sink_relations

    for source_sink_relation in relations_list.value:
        logger.debug(f"Parse source-sink relation '{source_sink_relation}'.")

        if not isinstance(source_sink_relation, model.RelationshipElement):
            logger.warning(f"'{source_sink_relation}' is not a RelationshipElement")
            continue

        if source_sink_relation.first is None or len(source_sink_relation.first.key) == 0:
            logger.warning(f"'first' reference is missing in RelationshipElement '{source_sink_relation.id_short}'")
            continue

        if source_sink_relation.second is None or len(source_sink_relation.second.key) == 0:
            logger.warning(f"'second' reference is missing in RelationshipElement '{source_sink_relation.id_short}'")
            continue

        relation = SourceSinkRelation()
        relation.source_properties = _get_reference_properties(source_sink_relation.first)
        relation.sink_properties = _get_reference_properties(source_sink_relation.second)

        source_sink_relations.append(relation)

    return source_sink_relations


def _get_reference_properties(reference: model.ExternalReference) -> ReferenceProperties | None:
    global_ref = next((key for key in reference.key if key.type == model.KeyTypes.GLOBAL_REFERENCE), None)

    if global_ref is None:
        logger.warning(f"No GLOBAL_REFERENCE key found in 'first' reference of RelationshipElement '{reference}'")
        return None

    last_fragment_ref = next((key for key in reversed(reference.key) if key.type == model.KeyTypes.FRAGMENT_REFERENCE), None)

    if last_fragment_ref is None:
        logger.warning(f"No FRAGMENT_REFERENCE key found in 'first' reference of RelationshipElement '{reference.id_short}'")
        return None

    parent_path = _get_reference_parent_path(reference)

    ref_properties = ReferenceProperties()
    ref_properties.submodel_id = global_ref.value
    ref_properties.property_name = last_fragment_ref.value
    ref_properties.parent_path = parent_path
    ref_properties.reference = reference

    return ref_properties


def _get_reference_parent_path(reference: model.ExternalReference) -> list[str]:
    """Get the parent path of a reference as a list of idShorts.

    :param reference: The reference to extract the parent path from.
    :return: A list of idShorts representing the parent path.
    """
    # Exclude the last key which is the actual element
    return [key.value for key in reference.key[:-1] if key.type == model.KeyTypes.FRAGMENT_REFERENCE]



"""Utility functions for version checking."""

import importlib.metadata
import logging

import requests

logger = logging.getLogger(__name__)


def check_for_update(package_name="aas-standard-parser"):
    """Check for updates of the package on PyPI.

    :param package_name: The name of the package to check for updates, defaults to "aas-standard-parser"
    """
    try:
        current_version = importlib.metadata.version(package_name)
        pypi_url = f"https://pypi.org/pypi/{package_name}/json"
        latest_version = requests.get(pypi_url, timeout=3).json()["info"]["version"]

        if current_version != latest_version:
            print(
                f"⚠️  A new version for package '{package_name}' is available: "
                f"{latest_version} (currently installed: {current_version}). "
                f"Use the following command to update the package: pip install --upgrade {package_name}"
            )
    except Exception as exc:
        logger.exception(f"Exception occurred while checking for package update: {exc}")



import pytest
from basyx.aas import model

from aas_standard_parser.descriptor_json_helper import parse_endpoint_href


def test_001a_get_mapping_configuration_root_element():
    href = "http://aas-server:8075/shells/aHR0cHM6Ly9mbHVpZDQwLmRlL2lkcy9zaGVsbC81NzkzXzU0NDlfNzgzMF80MjIz"

    href_data = parse_endpoint_href(href)

    assert href_data is not None
    assert href_data.base_url == "http://aas-server:8075"
    assert href_data.identifier == "aHR0cHM6Ly9mbHVpZDQwLmRlL2lkcy9zaGVsbC81NzkzXzU0NDlfNzgzMF80MjIz"
    assert href_data.identifier_decoded == "https://fluid40.de/ids/shell/5793_5449_7830_4223"

def test_001b_get_mapping_configuration_root_element():
    href = "http://aas-server:8075/submodels/RW5lcmd5TW9uaXRvcmluZw"

    href_data = parse_endpoint_href(href)

    assert href_data is not None
    assert href_data.base_url == "http://aas-server:8075"
    assert href_data.identifier == "RW5lcmd5TW9uaXRvcmluZw"
    assert href_data.identifier_decoded == "EnergyMonitoring"



"""Module for parsing submodels JSON data."""

import logging

logger = logging.getLogger(__name__)


def get_value_from_semantic_id_by_index(submodel_data: dict, index: int = 0) -> str | None:
    """Retrieve the value from the semantic ID from a submodel JSON dictionary by index.

    :param submodel_data: The submodel data as a dictionary.
    :param index: The index to access if the element is a list or collection, defaults to 0.
    :return: The value of the found submodel element or None if not found.
    """
    if "semanticId" not in submodel_data:
        logger.debug(f"Submodel '{submodel_data}' has no semantic ID")
        return None

    sm_semantic_id = submodel_data.get("semanticId", {})
    if "keys" not in sm_semantic_id or len(sm_semantic_id["keys"]) == 0:
        logger.debug(f"Submodel '{submodel_data}' has no semantic ID keys")
        return None

    keys = sm_semantic_id.get("keys", [])

    if len(keys) < index + 1:
        logger.debug(f"Submodel '{submodel_data}' has no semantic ID key at index {index}")
        return None

    key = keys[index]

    if "value" not in key:
        logger.debug(f"Submodel '{submodel_data}' has no semantic ID value")
        return None

    return key["value"]
